"""
This is inherited from pymatgen.io.cif
"""

from pymatgen.io.cif import CifParser
from pymatgen.core.operations import MagSymmOp, SymmOp
import warnings
from pymatgen.util.coord import find_in_coord_list_pbc, in_coord_list_pbc
from io import StringIO
import re
import numpy as np
from pymatgen.electronic_structure.core import Magmom
import math
from pymatgen.core.composition import Composition
from pymatgen.core.periodic_table import DummySpecies, Element, Species, get_el_sp
from pymatgen.symmetry.analyzer import SpacegroupOperations
from pymatgen.symmetry.structure import SymmetrizedStructure
from pymatgen.util.coord import find_in_coord_list_pbc
from kmcpy.external.structure import StructureKMCpy

class CifParserKMCpy(CifParser):

    @staticmethod
    def from_string(cif_string, **kwargs):
        """
        Creates a CifParser from a string.

        Args:
            cif_string (str): String representation of a CIF.
            **kwargs: Passthrough of all kwargs supported by CifParser.

        Returns:
            CifParser
        """
        stream = StringIO(cif_string)
        return CifParserKMCpy(stream, **kwargs)
    
    def _get_labeled_structure(self, data, primitive, symmetrized):
        """modified version from _get_structure, which can add the atom label to the site

        Args:
            data (dict): data parsed by the cif parser
            primitive (_type_): not implemented
            symmetrized (bool): not implemented
        """

        def get_num_implicit_hydrogens(sym):
            num_h = {"Wat": 2, "wat": 2, "O-H": 1}
            return num_h.get(sym[:3], 0)

        lattice = self.get_lattice(data)

        # if magCIF, get magnetic symmetry moments and magmoms
        # else standard CIF, and use empty magmom dict
        if self.feature_flags["magcif_incommensurate"]:
            raise NotImplementedError(
                "Incommensurate structures not currently supported."
            )
        if self.feature_flags["magcif"]:
            self.symmetry_operations = self.get_magsymops(data)
            magmoms = self.parse_magmoms(data, lattice=lattice)
        else:
            self.symmetry_operations = self.get_symops(data)
            magmoms = {}

        oxi_states = self._parse_oxi_states(data)

        coord_to_species = {}
        coord_to_magmoms = {}

        # add the coords_to_labels to record the label (Na1,Na2,Zr,S/P, etc.)
        coord_to_labels = {}

        def get_matching_coord(coord):
            keys = list(coord_to_species.keys())
            coords = np.array(keys)
            for op in self.symmetry_operations:
                c = op.operate(coord)
                inds = find_in_coord_list_pbc(coords, c, atol=self._site_tolerance)
                # can't use if inds, because python is dumb and np.array([0]) evaluates
                # to False
                if len(inds):
                    return keys[inds[0]]
            return False

        for i in range(len(data["_atom_site_label"])):

            try:
                # If site type symbol exists, use it. Otherwise, we use the
                # label.
                symbol = self._parse_symbol(data["_atom_site_type_symbol"][i])
                num_h = get_num_implicit_hydrogens(data["_atom_site_type_symbol"][i])
            except KeyError:
                symbol = self._parse_symbol(data["_atom_site_label"][i])
                num_h = get_num_implicit_hydrogens(data["_atom_site_label"][i])
            if not symbol:
                continue

            if oxi_states is not None:
                o_s = oxi_states.get(symbol, 0)
                # use _atom_site_type_symbol if possible for oxidation state
                if "_atom_site_type_symbol" in data.data.keys():
                    oxi_symbol = data["_atom_site_type_symbol"][i]
                    o_s = oxi_states.get(oxi_symbol, o_s)
                try:
                    el = Species(symbol, o_s)
                except Exception:
                    el = DummySpecies(symbol, o_s)
            else:
                el = get_el_sp(symbol)

            x = self.str2float(data["_atom_site_fract_x"][i])
            y = self.str2float(data["_atom_site_fract_y"][i])
            z = self.str2float(data["_atom_site_fract_z"][i])
            magmom = magmoms.get(data["_atom_site_label"][i], np.array([0, 0, 0]))

            try:
                occu = self.str2float(data["_atom_site_occupancy"][i])
            except (KeyError, ValueError):
                occu = 1

            if occu > 0:
                coord = (x, y, z)
                match = get_matching_coord(coord)
                comp_d = {el: occu}
                if num_h > 0:
                    comp_d["H"] = num_h
                    self.warnings.append(
                        "Structure has implicit hydrogens defined, "
                        "parsed structure unlikely to be suitable for use "
                        "in calculations unless hydrogens added."
                    )
                comp = Composition(comp_d)
                if not match:
                    coord_to_species[coord] = comp
                    coord_to_magmoms[coord] = magmom
                    coord_to_labels[coord] = data["_atom_site_label"][i]
                else:
                    coord_to_species[match] += comp
                    # disordered magnetic not currently supported
                    coord_to_magmoms[match] = None
                    coord_to_labels[coord] += data["_atom_site_label"][i]

        sum_occu = [
            sum(c.values())
            for c in coord_to_species.values()
            if not set(c.elements) == {Element("O"), Element("H")}
        ]
        if any(o > 1 for o in sum_occu):
            msg = (
                "Some occupancies ({}) sum to > 1! If they are within "
                "the occupancy_tolerance, they will be rescaled. "
                "The current occupancy_tolerance is set to: {}".format(
                    sum_occu, self._occupancy_tolerance
                )
            )
            warnings.warn(msg)
            self.warnings.append(msg)

        allspecies = []
        allcoords = []
        allmagmoms = []
        allhydrogens = []

        equivalent_indices = []

        site_properties = {}

        # initialize the wyckoff_sequence here
        site_properties["wyckoff_sequence"] = []

        site_properties["local_index"] = []

        site_properties["label"] = []

        local_index = 0

        # check to see if magCIF file is disordered
        if self.feature_flags["magcif"]:
            for k, v in coord_to_magmoms.items():
                if v is None:
                    # Proposed solution to this is to instead store magnetic
                    # moments as Species 'spin' property, instead of site
                    # property, but this introduces ambiguities for end user
                    # (such as unintended use of `spin` and Species will have
                    # fictitious oxidation state).
                    raise NotImplementedError(
                        "Disordered magnetic structures not currently supported."
                    )
        # print(coord_to_species)
        if coord_to_species.items():
            """
            for idx, (comp, group) in enumerate(
                    sorted(list(coord_to_species.items()), key=lambda x: x[1])
                ):
                print(comp,group)
            """

            for idx, (coord, comp) in enumerate(list(coord_to_species.items())):
                # I delete the weird group function
                # now just for every initial site, generate all sites

                # print(idx,comp)#debug

                tmp_coords = [coord]  # follow the fashion

                tmp_magmom = [coord_to_magmoms[tmp_coord] for tmp_coord in tmp_coords]
                tmp_label = [coord_to_labels[tmp_coord] for tmp_coord in tmp_coords]
                # print(tmp_coords,tmp_magmom,tmp_label)#debug

                if self.feature_flags["magcif"]:
                    coords, magmoms, _ = self._unique_coords(
                        tmp_coords, magmoms_in=tmp_magmom, lattice=lattice
                    )
                else:
                    coords, magmoms, _ = self._unique_coords(tmp_coords)

                # add wyckoff sequence here
                for wyckoff_sequence in range(0, len(coords)):
                    # tuple of wyckoff sequence, (label,sequence ) in format of  (str,int)

                    site_properties["wyckoff_sequence"].append(wyckoff_sequence)
                    site_properties["label"].append(coord_to_labels[tmp_coords[0]])
                    site_properties["local_index"].append(local_index)
                    local_index += 1

                if set(comp.elements) == {Element("O"), Element("H")}:
                    # O with implicit hydrogens
                    im_h = comp["H"]
                    species = Composition({"O": comp["O"]})
                else:
                    im_h = 0
                    species = comp

                # The following might be a more natural representation of equivalent indices,
                # but is not in the format expect by SymmetrizedStructure:
                #   equivalent_indices.append(list(range(len(allcoords), len(coords)+len(allcoords))))
                # The above gives a list like:
                #   [[0, 1, 2, 3], [4, 5, 6, 7, 8, 9, 10, 11]] where the
                # integers are site indices, whereas the version used below will give a version like:
                #   [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1]
                # which is a list in the same order as the sites, but where if a site has the same integer
                # it is equivalent.
                equivalent_indices += len(coords) * [idx]

                allhydrogens.extend(len(coords) * [im_h])
                allcoords.extend(coords)
                allspecies.extend(len(coords) * [species])
                allmagmoms.extend(magmoms)

            # rescale occupancies if necessary
            for i, species in enumerate(allspecies):
                totaloccu = sum(species.values())
                if 1 < totaloccu <= self._occupancy_tolerance:
                    allspecies[i] = species / totaloccu

        if (
            allspecies
            and len(allspecies) == len(allcoords)
            and len(allspecies) == len(allmagmoms)
        ):
            # site_properties = {}
            if any(allhydrogens):
                assert len(allhydrogens) == len(allcoords)
                site_properties["implicit_hydrogens"] = allhydrogens

            if self.feature_flags["magcif"]:
                site_properties["magmom"] = allmagmoms

            if len(site_properties) == 0:
                site_properties = None

            struct = StructureKMCpy(
                lattice, allspecies, allcoords, site_properties=site_properties
            )

            if symmetrized:

                # Wyckoff labels not currently parsed, note that not all CIFs will contain Wyckoff labels
                # TODO: extract Wyckoff labels (or other CIF attributes) and include as site_properties
                wyckoffs = ["Not Parsed"] * len(struct)

                # Names of space groups are likewise not parsed (again, not all CIFs will contain this information)
                # What is stored are the lists of symmetry operations used to generate the structure
                # TODO: ensure space group labels are stored if present
                sg = SpacegroupOperations("Not Parsed", -1, self.symmetry_operations)

                return SymmetrizedStructure(struct, sg, equivalent_indices, wyckoffs)

            struct = struct.get_sorted_structure()

            if primitive and self.feature_flags["magcif"]:
                struct = struct.get_primitive_structure(use_site_props=True)
            elif primitive:
                struct = struct.get_primitive_structure()
                struct = struct.get_reduced_structure()

            return struct

    def get_labeled_structures(self, primitive=True, symmetrized=False):
        """

        This is a modified version of get_structures for KMCPY, which preserve the label and the wyckoff sequence information.

        2022.03.30


        Return list of structures in CIF file. primitive boolean sets whether a
        conventional cell structure or primitive cell structure is returned.

        Args:
            primitive (bool): Set to False to return conventional unit cells.
                Defaults to True. With magnetic CIF files, will return primitive
                magnetic cell which may be larger than nuclear primitive cell.
            symmetrized (bool): If True, return a SymmetrizedStructure which will
                include the equivalent indices and symmetry operations used to
                create the Structure as provided by the CIF (if explicit symmetry
                operations are included in the CIF) or generated from information
                in the CIF (if only space group labels are provided). Note that
                currently Wyckoff labels and space group labels or numbers are
                not included in the generated SymmetrizedStructure, these will be
                notated as "Not Parsed" or -1 respectively.

        Returns:
            List of Structures.
        """

        if primitive and symmetrized:
            raise ValueError(
                "Using both 'primitive' and 'symmetrized' arguments is not currently supported "
                "since unexpected behavior might result."
            )

        structures = []
        for i, d in enumerate(self._cif.data.values()):
            try:
                s = self._get_labeled_structure(d, primitive, symmetrized)
                if s:
                    structures.append(s)
            except (KeyError, ValueError) as exc:
                # Warn the user (Errors should never pass silently)
                # A user reported a problem with cif files produced by Avogadro
                # in which the atomic coordinates are in Cartesian coords.
                self.warnings.append(str(exc))
                warnings.warn(
                    f"No structure parsed for {i + 1} structure in CIF. Section of CIF file below."
                )
                warnings.warn(str(d))
                warnings.warn(f"Error is {str(exc)}.")

        if self.warnings:
            warnings.warn(
                "Issues encountered while parsing CIF: %s" % "\n".join(self.warnings)
            )
        if len(structures) == 0:
            raise ValueError("Invalid cif file with no structures!")
        return structures

    def _unique_coords(
        self,
        coords,
        magmoms=None,
        lattice=None,
        labels=None,
    ):
        """
        Generate unique coordinates using coord and symmetry positions
        and also their corresponding magnetic moments, if supplied.
        """
        coords_out: list[np.ndarray] = []
        labels_out = []
        labels = labels or {}

        if magmoms:
            magmoms_out = []
            if len(magmoms) != len(coords):
                raise ValueError
            for tmp_coord, tmp_magmom in zip(coords, magmoms):
                for op in self.symmetry_operations:
                    coord = op.operate(tmp_coord)
                    coord = np.array([i - math.floor(i) for i in coord])
                    if isinstance(op, MagSymmOp):
                        # Up to this point, magmoms have been defined relative
                        # to crystal axis. Now convert to Cartesian and into
                        # a Magmom object.
                        magmom = Magmom.from_moment_relative_to_crystal_axes(
                            op.operate_magmom(tmp_magmom), lattice=lattice
                        )
                    else:
                        magmom = Magmom(tmp_magmom)
                    if not in_coord_list_pbc(
                        coords_out, coord, atol=self._site_tolerance
                    ):
                        coords_out.append(coord)
                        magmoms_out.append(magmom)
                        labels_out.append(labels.get(tmp_coord))
            return coords_out, magmoms_out, labels_out

        for tmp_coord in coords:
            for op in self.symmetry_operations:
                coord = op.operate(tmp_coord)
                coord = np.array([i - math.floor(i) for i in coord])
                if not in_coord_list_pbc(coords_out, coord, atol=self._site_tolerance):
                    coords_out.append(coord)
                    labels_out.append(labels.get(tmp_coord))

        dummy_magmoms = [Magmom(0)] * len(coords_out)
        return coords_out, dummy_magmoms, labels_out

    @staticmethod
    def str2float(text):
        """
        Remove uncertainty brackets from strings and return the float.
        """

        try:
            # Note that the ending ) is sometimes missing. That is why the code has
            # been modified to treat it as optional. Same logic applies to lists.
            return float(re.sub(r"\(.+\)*", "", text))
        except TypeError:
            if isinstance(text, list) and len(text) == 1:
                return float(re.sub(r"\(.+\)*", "", text[0]))
        except ValueError as ex:
            if text.strip() == ".":
                return 0
            raise ex
        raise ValueError(f"{text} cannot be converted to float")



#!/usr/bin/env python
"""
This module provides the KMC class and associated functions for performing Kinetic Monte Carlo (kMC) simulations, particularly for modeling stochastic processes in materials such as ion diffusion. The KMC class manages the initialization, event handling, probability calculations, and simulation loop for kMC workflows. It supports loading input data from various sources, updating system states, and tracking simulation results.
"""
from numba.typed import List
from numba import njit
from kmcpy.external.structure import StructureKMCpy
import numpy as np
import pandas as pd
from copy import copy
import json
from kmcpy.model import LocalClusterExpansion
from kmcpy.tracker import Tracker
from kmcpy.event import Event
from kmcpy.io import convert
import logging
import kmcpy
from kmcpy.io import InputSet

logger = logging.getLogger(__name__) 
logging.getLogger('numba').setLevel(logging.WARNING)

class KMC:
    """Kinetic Monte Carlo Simulation Class.

    This class implements a Kinetic Monte Carlo (kMC) simulation for modeling
    stochastic processes in materials, such as ion diffusion. It provides
    methods for initializing the simulation from various input sources,
    managing events, updating system states, and running the simulation loop.
    """
    def __init__(self, initial_occ: list,
                supercell_shape: list,
                fitting_results: str,
                fitting_results_site: str,
                lce_fname: str,
                lce_site_fname: str,
                template_structure_fname: str,
                event_fname: str,
                event_kernel: str,
                v: float = 1e13,
                temperature: float = 300,
                convert_to_primitive_cell: bool=False,
                immutable_sites: list=[],**kwargs)-> None:
        """Initialize the Kinetic Monte Carlo (kMC) simulation.

        It is recommended to use the `from_inputset` method to initialize the
        KMC object from an InputSet.

        Args:
            initial_occ (list): The initial occupation list representing the
            configuration of the system.
            supercell_shape (list): Shape of the supercell as a list of
            integers (e.g., [2, 2, 2]). This should be consistent with
            events.
            fitting_results (str): Path to the JSON file containing the
            fitting results for E_kra.
            fitting_results_site (str): Path to the JSON file containing the
            fitting results for site energy difference.
            lce_fname (str): Path to the JSON file containing the Local
            Cluster Expansion (LCE) model.
            lce_site_fname (str): Path to the JSON file containing the site
            LCE model for computing site energy differences.
            template_structure_fname (str): Path to the CIF file of the
            template structure (with all sites filled).
            event_fname (str): Path to the JSON file containing the list of
            events.
            event_kernel (str): Path to the event kernel file.
            v (float, optional): Attempt frequency (prefactor) for hopping
            events. Defaults to 1e13 Hz.
            temperature (float, optional): Simulation temperature in Kelvin.
            Defaults to 300 K.
            convert_to_primitive_cell (bool, optional): Whether to convert the
            structure to its primitive cell (default: False).
            immutable_sites (list, optional): List of sites to be treated as
            immutable and removed from the simulation (default: []).

        Notes:
            - Loads structure, fitting results, LCE models, and events.
            - Initializes the supercell, occupation, event list, and hopping
              probabilities.
            - Logs detailed information about the initialization process.

        """
        logger.info(kmcpy.get_logo())
        logger.info(f"Initializing kMC calculations ...")
        self.structure = StructureKMCpy.from_cif(
            template_structure_fname, primitive=convert_to_primitive_cell
        )
        supercell_shape_matrix = np.diag(supercell_shape)
        logger.info(f"Supercell Shape:\n {supercell_shape_matrix}")

        logger.info("Converting to the supercell ...")
        logger.debug("removing the immutable sites: {immutable_sites}")
        self.structure.remove_species(immutable_sites)
        self.structure.make_supercell(supercell_shape_matrix)

        logger.info("Loading fitting results: E_kra ...")
        fitting_results = (
            pd.read_json(fitting_results, orient="index")
            .sort_values(by=["time_stamp"], ascending=False)
            .iloc[0]
        )
        self.keci = fitting_results.keci
        self.empty_cluster = fitting_results.empty_cluster

        logger.info("Loading fitting results: site energy ...")
        fitting_results_site = (
            pd.read_json(fitting_results_site, orient="index")
            .sort_values(by=["time_stamp"], ascending=False)
            .iloc[0]
        )
        self.keci_site = fitting_results_site.keci
        self.empty_cluster_site = fitting_results_site.empty_cluster

        logger.info(f"Loading occupation: {initial_occ}")
        self.occ_global = copy(initial_occ)

        logger.info(f"Loading LCE models from: {lce_fname} and {lce_site_fname}")
        local_cluster_expansion = LocalClusterExpansion.from_json(lce_fname)
        local_cluster_expansion_site = LocalClusterExpansion.from_json(lce_site_fname)

        # sublattice_indices are the orbits in table S3 of support information
        # Convert them into Numba List for faster execution
        sublattice_indices = _convert_list(local_cluster_expansion.sublattice_indices)
        sublattice_indices_site = _convert_list(
            local_cluster_expansion_site.sublattice_indices
        )
        events_site_list = []

        logger.info(f"Loading events from: {event_fname}")
        with open(event_fname, "rb") as fhandle:
            events_dict = json.load(fhandle)

        logger.info("Initializing correlation matrix and E_kra for all events ...")
        events = []
        for event_dict in events_dict:
            event = Event.from_dict(event_dict)
            event.set_sublattice_indices(sublattice_indices, sublattice_indices_site)
            event.initialize_corr()
            event.update_event(
            self.occ_global,
            v,
            temperature,
            self.keci,
            self.empty_cluster,
            self.keci_site,
            self.empty_cluster_site,
            )
            events_site_list.append(event.local_env_indices_list)
            events.append(event)

        self.events = events

        logger.info("Initializing hopping probabilities ...")
        # Preallocate prob_list and prob_cum_list arrays for reuse
        self.prob_list = np.empty(len(events), dtype=np.float64)
        for i, event in enumerate(events):
            self.prob_list[i] = event.probability
        self.prob_cum_list = np.empty(len(events), dtype=np.float64)
        np.cumsum(self.prob_list, out=self.prob_cum_list)

        logger.info("Fitted time and error (LOOCV,RMS)")
        logger.info(f"{fitting_results.time}, {fitting_results.loocv}, {fitting_results.rmse}")
        logger.info("Fitted KECI and empty cluster E_kra")
        logger.info(f"{self.keci}, {self.empty_cluster}")

        logger.info("Fitted time and error (LOOCV,RMS) (site energy)")
        logger.info(
            f"{fitting_results_site.time}, {fitting_results_site.loocv}, {fitting_results_site.rmse}"
        )
        logger.info("Fitted KECI and empty cluster (site energy or E_end)")
        logger.info(f"{self.keci_site}, {self.empty_cluster_site}")
        logger.info(f"Lists for each event {events_site_list}")
        logger.info(f"Hopping probabilities: {self.prob_list}")
        logger.info(f"Cumulative sum of hopping probabilities: {self.prob_cum_list}")
        logger.info(f"Loading event kernel from: {event_kernel}")
        self.load_site_event_list(fname = event_kernel)

    @classmethod
    def from_inputset(cls, inputset: InputSet)-> "KMC":
        """Initialize KMC from InputSet object.

        Args:
            inputset (kmcpy.io.InputSet): InputSet object containing all
            necessary parameters.

        Returns:
            KMC: An instance of the KMC class initialized with parameters
            from the InputSet.
        """
        params = {k: v for k, v in inputset._parameters.items() if k != "task"}
        return cls(**params)

    def load_site_event_list(
        self, fname:str="../event_kernal.csv"
    )->None:  # workout the site_event_list -> site_event_list[site_index] will return a list of event index to update if a site_index is chosen
        logger.info("Working at the site_event_list ...")

        logger.info(f"Loading {fname}")
        site_event_list = []
        with open(fname) as f:
            data = f.readlines()
        for x in data:
            if len(x.strip()) == 0:
                site_event_list.append([])
            else:
                site_event_list.append([int(y) for y in x.strip().split()])
        self.site_event_list = site_event_list

    def show_project_info(self):
        try:
            logger.info("Probabilities:")
            logger.info(self.prob_list)
            logger.info("Cumultative probability list:")
            logger.info(self.prob_cum_list / sum(self.prob_list))
        except Exception:
            pass

    def propose(
        self,
        events: list,
    ) -> tuple[Event, float]:
        """Propose a new event to be updated by update().

        Args:
            events (list): List of events.

        Returns:
            tuple[Event, float]: The chosen event and the time for this event to occur.
        """
        proposed_event_index, dt = _propose(prob_cum_list=self.prob_cum_list, rng=self.rng)
        event = events[proposed_event_index]
        return event, dt
    

    def update(self, event: Event, events: list[Event])-> None:
        """
        Updates the system state and event probabilities after an event occurs.
        This method performs the following steps:
        1. Flips the occupation values of the two mobile ion species involved in the event.
        2. Identifies all events that need to be updated due to the change in occupation.
        3. Updates each affected event's properties and recalculates their probabilities.
        4. Updates the cumulative probability list for event selection.

        Args:
            event: The event object that has just occurred, containing indices of the affected mobile ion species.
            events (list): List of all event objects in the system.
            
        Side Effects:
            Modifies `self.occ_global`, `self.prob_list`, and `self.prob_cum_list` in place.
        """

        self.occ_global[event.mobile_ion_specie_1_index] *= -1
        self.occ_global[event.mobile_ion_specie_2_index] *= -1
        events_to_be_updated = copy(
            self.site_event_list[event.mobile_ion_specie_2_index]
        )  # event_to_be_updated= list, include the indices, of the event that need to be updated.
        for e_index in events_to_be_updated:
            events[e_index].update_event(
                self.occ_global,
                self.inputset.v,
                self.inputset.temperature,
                self.keci,
                self.empty_cluster,
                self.keci_site,
                self.empty_cluster_site,
            )
            self.prob_list[e_index] = copy(events[e_index].probability)
        self.prob_cum_list = np.cumsum(self.prob_list)

    def run(self, inputset: InputSet, label: str = None)-> Tracker:
        """Run KMC simulation from an InputSet object.

        Args:
            inputset (InputSet): InputSet object containing all necessary parameters.
            label (str, optional): Label for the simulation run. Defaults to None.

        Returns:
            kmcpy.tracker.Tracker: Tracker object containing simulation results.
        """
        self.inputset = inputset
        if hasattr(inputset, "random_seed") and inputset.random_seed is not None:
            self.rng = np.random.default_rng(seed=inputset.random_seed)
        else:
            self.rng = np.random.default_rng()

        logger.info(f"Simulation condition: v = {inputset.v} T = {inputset.temperature}")
        pass_length = len(
            [
            el.symbol
            for el in self.structure.species
            if inputset.mobile_ion_specie in el.symbol
            ]
        )
        logger.info("============================================================")
        logger.info("Start running kMC ... ")
        logger.info("Initial occ_global, prob_list and prob_cum_list")
        logger.info("Starting Equilbrium ...")
        for _ in np.arange(inputset.equ_pass):
            for _ in np.arange(pass_length):
                event, dt = self.propose(self.events)
                self.update(event, self.events)

        logger.info("Start running kMC ...")

        tracker = Tracker.from_inputset(inputset = inputset,
                                        structure = self.structure,
                                        occ_initial = self.occ_global)
        
        logger.info(
            "Pass\tTime\t\tMSD\t\tD_J\t\tD_tracer\tConductivity\tH_R\t\tf"
        )
        for current_pass in np.arange(inputset.kmc_pass):
            for _ in np.arange(pass_length):
                event, dt = self.propose(self.events)
                tracker.update(event, self.occ_global, dt)
                self.update(event, self.events)
            
            tracker.update_current_pass(current_pass)
            tracker.compute_properties()
            tracker.show_current_info()

        tracker.write_results(self.occ_global, label=label)
        return tracker

    def as_dict(self)-> dict:
        d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "structure": self.structure.as_dict(),
            "keci": self.keci,
            "empty_cluster": self.empty_cluster,
            "keci_site": self.keci_site,
            "empty_cluster_site": self.empty_cluster_site,
            "occ_global": self.occ_global,
        }
        return d

    def to_json(self, fname)-> None:
        logger.info(f"Saving: {fname}")
        with open(fname, "w") as fhandle:
            d = self.as_dict()
            jsonStr = json.dumps(
                d, indent=4, default=convert
            )  # to get rid of errors of int64
            fhandle.write(jsonStr)

    @classmethod
    def from_json(cls, fname)-> "KMC":
        logger.info(f"Loading: {fname}")
        with open(fname, "rb") as fhandle:
            objDict = json.load(fhandle)
        obj = KMC()
        obj.__dict__ = objDict
        logger.info("load complete")
        return obj


def _convert_list(list_to_convert)-> List:
    converted_list = List([List(List(j) for j in i) for i in list_to_convert])
    return converted_list


@njit
def _propose(prob_cum_list, rng)-> tuple[int, float]:
    random_seed = rng.random()
    random_seed_2 = rng.random()
    proposed_event_index = np.searchsorted(
        prob_cum_list / prob_cum_list[-1], random_seed, side="right"
    )
    dt = (-1.0 / prob_cum_list[-1]) * np.log(random_seed_2)
    return proposed_event_index, dt


import unittest
import pytest
import os

current_dir = os.path.dirname(os.path.abspath(__file__))
file_path = os.path.join(current_dir, 'files')

class TestNASICONbulk(unittest.TestCase):

    @pytest.mark.order(1)
    def test_neighbor_info_matcher(self):
        print("neighbor info matcher testing")
        
        from kmcpy.event_generator import neighbor_info_matcher
        import numpy as np

        from kmcpy.external.structure import StructureKMCpy
        from kmcpy.external.local_env import CutOffDictNNKMCpy

        nasicon = StructureKMCpy.from_cif(
            f"{file_path}/EntryWithCollCode15546_Na4Zr2Si3O12_573K.cif", primitive=True
        )

        center_Na1 = [0, 1]
        local_env_finder = CutOffDictNNKMCpy({("Na+", "Na+"): 4, ("Na+", "Si4+"): 4})

        reference_neighbor_sequences = sorted(
            sorted(
                local_env_finder.get_nn_info(nasicon, center_Na1[0]),
                key=lambda x: x["wyckoff_sequence"],
            ),
            key=lambda x: x["label"],
        )

        np.set_printoptions(precision=2, suppress=True)
        reference_neighbor = neighbor_info_matcher.from_neighbor_sequences(
            neighbor_sequences=reference_neighbor_sequences
        )

        print(reference_neighbor.neighbor_species)
        print(reference_neighbor.distance_matrix)

        wrong_sequence_neighbor = sorted(
            sorted(
                local_env_finder.get_nn_info(nasicon, center_Na1[1]),
                key=lambda x: x["wyckoff_sequence"],
            ),
            key=lambda x: x["label"],
        )

        self.assertFalse(
            np.allclose(
                neighbor_info_matcher.from_neighbor_sequences(
                    neighbor_sequences=wrong_sequence_neighbor
                ).distance_matrix,
                reference_neighbor.distance_matrix,
                rtol=0.01,
                atol=0.01,
            )
        )

        resorted_wrong_sequence = reference_neighbor.brutal_match(
            wrong_sequence_neighbor, rtol=0.01
        )

        resorted_neighbor = neighbor_info_matcher.from_neighbor_sequences(
            neighbor_sequences=resorted_wrong_sequence
        )

        self.assertTrue(
            np.allclose(
                reference_neighbor.distance_matrix,
                resorted_neighbor.distance_matrix,
                rtol=0.01,
                atol=0.01,
            )
        )

    @pytest.mark.order(2)
    def test_generate_events(self):
        mobile_ion_identifier_type = "label"
        mobile_ion_specie_1_identifier = "Na1"
        mobile_ion_specie_2_identifier = "Na2"
        template_structure_fname = f"{file_path}/EntryWithCollCode15546_Na4Zr2Si3O12_573K.cif"
        local_env_cutoff_dict = {("Na+", "Na+"): 4, ("Na+", "Si4+"): 4}
        from kmcpy.event_generator import generate_events

        generate_events(
            template_structure_fname=template_structure_fname,
            local_env_cutoff_dict=local_env_cutoff_dict,
            mobile_ion_identifier_type=mobile_ion_identifier_type,
            mobile_ion_specie_1_identifier=mobile_ion_specie_1_identifier,
            mobile_ion_specie_2_identifier=mobile_ion_specie_2_identifier,
            species_to_be_removed=["O2-", "O", "Zr4+", "Zr"],
            distance_matrix_rtol=0.01,
            distance_matrix_atol=0.01,
            find_nearest_if_fail=False,
            convert_to_primitive_cell=False,
            export_local_env_structure=True,
            supercell_shape=[2, 1, 1],
            event_fname=f"{file_path}/events.json",
            event_kernal_fname=f"{file_path}/event_kernal.csv",
        )

        reference_local_env_dict = generate_events(
            template_structure_fname=template_structure_fname,
            local_env_cutoff_dict=local_env_cutoff_dict,
            mobile_ion_identifier_type=mobile_ion_identifier_type,
            mobile_ion_specie_1_identifier=mobile_ion_specie_1_identifier,
            mobile_ion_specie_2_identifier=mobile_ion_specie_2_identifier,
            species_to_be_removed=["O2-", "O", "Zr4+", "Zr"],
            distance_matrix_rtol=0.01,
            distance_matrix_atol=0.01,
            find_nearest_if_fail=False,
            convert_to_primitive_cell=True,
            export_local_env_structure=True,
            supercell_shape=[2, 1, 1],
            event_fname=f"{file_path}/events.json",
            event_kernal_fname=f"{file_path}/event_kernal.csv",
        )

        print("reference_local_env_dict:", reference_local_env_dict)

        self.assertEqual(
            len(reference_local_env_dict), 1
        )  # only one type of local environment should be found. If more than 1, raise error.

    @pytest.mark.order(3)
    def test_generate_local_cluster_exapnsion(self):
        
        from kmcpy.model import LocalClusterExpansion

        mobile_ion_identifier_type = "label"
        mobile_ion_specie_1_identifier = "Na1"
        a = LocalClusterExpansion()
        a.initialization(
            mobile_ion_identifier_type=mobile_ion_identifier_type,
            mobile_ion_specie_1_identifier=mobile_ion_specie_1_identifier,
            cutoff_cluster=[6, 6, 0],
            cutoff_region=4,
            template_structure_fname=f"{file_path}/EntryWithCollCode15546_Na4Zr2Si3O12_573K.cif",
            convert_to_primitive_cell=True,
        )
        a.to_json(f"{file_path}/input/lce.json")
        self.assertEqual(1, 1)

    def test_fitting(self):
        from kmcpy.fitting import Fitting
        import numpy as np

        local_cluster_expansion_fit = Fitting()

        y_pred, y_true = local_cluster_expansion_fit.fit(
            alpha=1.5,
            max_iter=1000000,
            ekra_fname=f"{file_path}/fitting/local_cluster_expansion/e_kra.txt",
            keci_fname=f"{file_path}/keci.txt",
            weight_fname=f"{file_path}/fitting/local_cluster_expansion/weight.txt",
            corr_fname=f"{file_path}/fitting/local_cluster_expansion/correlation_matrix.txt",
            fit_results_fname=f"{file_path}/fitting_results.json",
        )
        print("fitting", y_pred, y_true)
        self.assertTrue(np.allclose(y_pred, y_true, rtol=0.3, atol=10.0))

    @pytest.mark.order(4)
    def test_kmc_main_function(self):
        from kmcpy.io import InputSet
        from kmcpy.kmc import KMC
        import numpy as np

        inputset = InputSet.from_json(f"{file_path}/input/kmc_input.json")

        print(inputset._parameters.keys())
        print(inputset._parameters["initial_state"])

        kmc = KMC.from_inputset(inputset)

        kmc_tracker = kmc.run(inputset)

        print(kmc_tracker.return_current_info())
        self.assertTrue(
            np.allclose(
                np.array(kmc_tracker.return_current_info()),
                np.array(
                    (1.1193006038758543e-06, 307.37444494263616, 1.4630573145769372e-08, 4.5768825621743376e-09, 1.1823906621661553, 0.312830024946617, 0.21998150220477225)
                ),
                rtol=0.01,
                atol=0.01,
            )
        )

        # np.array((3.517242770690013e-06, 26.978226076495748, 3.187544456106211e-10, 1.2783794881088614e-10, 0.025760595723683707, 0.4010546380490277, 0.04309185078659044)) this is run from the given random number kernal and random number seed. This is a very strict criteria to see if the behavior of KMC is correct
        # with 0-7, 32-37 selected: np.array(1.1193006038758543e-06, 307.37444494263616, 1.4630573145769372e-08, 4.5768825621743376e-09, 1.1823906621661553, 0.312830024946617, 0.21998150220477225)

    @pytest.mark.order(5)
    def test_kmc_main_function_randomized(self):
        from kmcpy.io import InputSet
        from kmcpy.kmc import KMC
        import numpy as np

        inputset = InputSet.from_json(f"{file_path}/input/kmc_input.json")

        print(inputset._parameters.keys())
        print(inputset._parameters["initial_state"])
        
        inputset.set_parameter("random_seed", np.random.randint(0, 1000000))
        kmc = KMC.from_inputset(inputset)
        
        kmc_tracker = kmc.run(inputset)

        print(kmc_tracker.return_current_info())
        self.assertFalse(
            np.allclose(
                np.array(kmc_tracker.return_current_info()),
                np.array(
                    (1.1193006038758543e-06, 307.37444494263616, 1.4630573145769372e-08, 4.5768825621743376e-09, 1.1823906621661553, 0.312830024946617, 0.21998150220477225)
                ),
                rtol=0.01,
                atol=0.01,
            )
        )

        # np.array((3.517242770690013e-06, 26.978226076495748, 3.187544456106211e-10, 1.2783794881088614e-10, 0.025760595723683707, 0.4010546380490277, 0.04309185078659044)) this is run from the given random number kernal and random number seed. This is a very strict criteria to see if the behavior of KMC is correct
        # with 0-7, 32-37 selected: np.array(1.1193006038758543e-06, 307.37444494263616, 1.4630573145769372e-08, 4.5768825621743376e-09, 1.1823906621661553, 0.312830024946617, 0.21998150220477225)
    def test_gather_mc_data(self):
        
        from kmcpy.tools.gather_mc_data import generate_supercell, gather_data
        from kmcpy.external.structure import StructureKMCpy
        import numpy as np

        structure_from_json = generate_supercell(
            f"{file_path}/gather_mc_data/prim.json", (8, 8, 8)
        )
        df = gather_data(f"{file_path}/gather_mc_data/comp*", structure_from_json)
        df.to_json(f"{file_path}/gather_mc_data/mc_results_json.json", orient="index")
        occ1 = df["occ"]

        structure_from_cif = StructureKMCpy.from_cif(
            f"{file_path}/EntryWithCollCode15546_Na4Zr2Si3O12_573K.cif", primitive=True
        )
        structure_from_cif.remove_species(["Zr", "O", "Zr4+", "O2-"])
        structure_from_cif.remove_oxidation_states()
        structure_from_cif = structure_from_cif.make_kmc_supercell([8, 8, 8])
        df2 = gather_data(f"{file_path}/gather_mc_data/comp*", structure_from_cif)
        df2.to_json(f"{file_path}/gather_mc_data/mc_results_cif.json", orient="index")
        occ2 = df2["occ"]
        for i in range(0, len(occ1[0])):
            if occ1[0][i] != occ2[0][i]:
                print(i, occ1[i], occ2[i])
        self.assertTrue(np.allclose(occ1[0], occ2[0], rtol=0.001, atol=0.001))


if __name__ == "__main__":
    unittest.main()



import unittest
import os

current_dir = os.path.dirname(os.path.abspath(__file__))
file_path = os.path.join(current_dir, 'files')

class TestNa3SbS4(unittest.TestCase):

    def test_generate_events(self):
        from pathlib import Path
        import os

        current_dir = Path(__file__).absolute().parent
        os.chdir(current_dir)
        mobile_ion_identifier_type = "label"
        mobile_ion_specie_1_identifier = "Na1"
        mobile_ion_specie_2_identifier = "Na1"
        template_structure_fname = f"{file_path}/Na3SbS4_cubic.cif"
        local_env_cutoff_dict = {("Na+", "Na+"): 5, ("Na+", "Sb5+"): 4}
        from kmcpy.event_generator import generate_events

        reference_local_env_dict = generate_events(
            template_structure_fname=template_structure_fname,
            local_env_cutoff_dict=local_env_cutoff_dict,
            mobile_ion_identifier_type=mobile_ion_identifier_type,
            mobile_ion_specie_1_identifier=mobile_ion_specie_1_identifier,
            mobile_ion_specie_2_identifier=mobile_ion_specie_2_identifier,
            species_to_be_removed=["S2-", "S", "Zr4+", "Zr"],
            distance_matrix_rtol=0.01,
            distance_matrix_atol=0.01,
            find_nearest_if_fail=False,
            convert_to_primitive_cell=False,
            export_local_env_structure=True,
            supercell_shape=[3, 3, 3],
            event_fname=f"{file_path}/events.json",
            event_kernal_fname=f"{file_path}/event_kernal.csv",
        )

        self.assertEqual(
            len(reference_local_env_dict), 1
        )  # only one type of local environment should be found. If more than 1, raise error.

    def test_generate_local_cluster_exapnsion(self):
        from pathlib import Path
        import os

        current_dir = Path(__file__).absolute().parent
        os.chdir(current_dir)
        from kmcpy.model import LocalClusterExpansion

        mobile_ion_identifier_type = "label"
        mobile_ion_specie_1_identifier = "Na1"
        a = LocalClusterExpansion()
        a.initialization(
            mobile_ion_identifier_type=mobile_ion_identifier_type,
            mobile_ion_specie_1_identifier=mobile_ion_specie_1_identifier,
            cutoff_cluster=[6, 6, 0],
            cutoff_region=5,
            template_structure_fname=f"{file_path}/Na3SbS4_cubic.cif",
            convert_to_primitive_cell=True,
            is_write_basis=True,
            species_to_be_removed=["S2-", "S"],
        )
        self.assertEqual(1, 1)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
import numpy as np
import pandas as pd
import glob2, json
from pymatgen.core.lattice import Lattice
import numba as nb
from numba.typed import List
from joblib import Parallel, delayed
import multiprocessing, os
from kmcpy.external.structure import StructureKMCpy


def generate_supercell(prim_fname, supercell_shape):
    shape = supercell_shape
    print("Initializing model with pirm.json at", prim_fname, "...")
    with open(prim_fname, "r") as f:
        prim = json.load(f)
        prim_coords = [site["coordinate"] for site in prim["basis"]]
        prim_specie_cases = [site["occupant_dof"] for site in prim["basis"]]
        prim_lattice = Lattice(prim["lattice_vectors"])
        prim_species = [s[0] for s in prim_specie_cases]

    supercell_shape_matrix = np.diag(supercell_shape)
    print("Supercell Shape:\n", supercell_shape_matrix)
    structure = StructureKMCpy(prim_lattice, prim_species, prim_coords)
    print("Converting supercell ...")
    structure.remove_species(["Zr", "O"])
    structure.make_supercell(supercell_shape_matrix)
    # structure.to(fmt='cif',filename='supercell.cif')
    return structure


def gather_data(path, template_structure):
    locations = glob2.glob(path)
    print(locations)
    get_occ(locations[0] + "/conditions.0/trajectory/POSCAR.final", template_structure)
    occ = Parallel(n_jobs=multiprocessing.cpu_count())(
        delayed(get_occ)(
            location + "/conditions.0/trajectory/POSCAR.final", template_structure
        )
        for location in locations
    )
    data = []
    for index, location in enumerate(locations):
        print('location',location)
        current_path = os.path.split(location)[-1]
        x = float(current_path.split("_")[1])
        structure_index = int(current_path.split("_")[2])
        print(location, occ[index])
        data.append([x, structure_index, occ[index]])
    return pd.DataFrame(data, columns=["comp", "structure_index", "occ"]).sort_values(
        by=["comp", "structure_index"]
    )


def isExist(s, structure):
    is_exist = False
    for site_in_structure in structure.sites:
        if s.species == site_in_structure.species:
            if site_in_structure.distance(s) < 1e-3:
                is_exist = True
                break
    return is_exist


"""
get_occ() works out the occupation vector by comparing the POSCAR generated from the Monte Carlo with the template structure using following steps:
1. get frac_coords and species of each sites in casm_structure and template_structure
2. take frac_coords difference by considering periodic boundary condition
3. for those norm(difference) = 0 (should be only one site), compare their species
4. if norm(difference) ==0 and species matches, then occ=-1 otherwise occ=1
"""


def get_occ(mc_poscar, template_structure):
    print(mc_poscar)
    casm_structure = StructureKMCpy.from_file(mc_poscar)
    casm_structure.remove_species(["Zr", "O"])
    occ = []
    template_frac_coords = np.array(template_structure.frac_coords)
    template_species = List([i.symbol for i in template_structure.species])
    casm_frac_coords = np.array(casm_structure.frac_coords)
    casm_species = List([i.symbol for i in casm_structure.species])
    lattice = template_structure.lattice.matrix

    occ = _get_occ(
        casm_frac_coords, template_frac_coords, casm_species, template_species, lattice
    )
    print(occ)
    print(len(np.where(occ == -1)[0]))

    print("Finished", mc_poscar)
    return np.array(occ)


@nb.njit
def _get_occ(
    casm_frac_coords, template_frac_coords, casm_species, template_species, lattice
):
    occ = np.zeros(len(template_species), dtype=np.int64)
    for i, (template_frac_coord, template_specie) in enumerate(
        zip(template_frac_coords, template_species)
    ):

        diff_frac = casm_frac_coords - template_frac_coord
        diff_frac = diff_frac - rnd(diff_frac)  # for periodic condition
        diff_cart = diff_frac @ lattice
        dists_cart = norm(diff_cart)
        matched_site_idx = np.where(dists_cart < 1e-6)[0]
        if (
            matched_site_idx.size > 0
            and casm_species[matched_site_idx[0]] in template_specie
        ):
            occ[i] = -1
        else:
            print(matched_site_idx.size)
            occ[i] = 1
        # else:
        #     print(i,'is a Vacancy')
    return occ


@nb.njit
def rnd(x):
    out = np.empty_like(x)
    np.around(x, 0, out)
    return out


@nb.njit
def norm(x):
    out = np.zeros(len(x))
    for j, i in enumerate(x):
        out[j] = (i[0] ** 2 + i[1] ** 2 + i[2] ** 2) ** 0.5
    return out



import importlib
import pytest

@pytest.mark.parametrize("module_path", [
    "kmcpy.external.structure",
    "kmcpy.external.cif",
    "kmcpy.external.local_env",
])
def test_module_imports(module_path):
    """Test that modules can be imported without circular import errors."""
    try:
        importlib.import_module(module_path)
    except ImportError as e:
        pytest.fail(f"Failed to import {module_path}: {e}")

@pytest.mark.parametrize("module_path", [
    "kmcpy.external",
])
def test_top_level_imports(module_path):
    try:
        importlib.import_module(module_path)
    except ImportError as e:
        pytest.fail(f"Failed to import top-level {module_path}: {e}")



#!/usr/bin/env python
import numpy as np
import pandas as pd
import glob2, os, json

"""
KMC data gathering
Here are the format for each results file
self.results = {'D_J':[],'D_tracer':[],'conductivity':[],'f':[],'H_R':[],
        'occupation':[],'ekra':[],'displacement':[],'hop_count':[]}

shape of matrices
hop_count[n_pass,n_na]
displacement[n_pass,n_na,3]
ekra[n_pass]
"""


def search_string_list(fname, word, location):
    res = []
    with open(fname) as f:
        for l in f.readlines():
            if word in l:
                res.append(float(l.strip().split()[location]))
    return res


def gather_data(path):
    locations = glob2.glob(path)
    data = []
    print(locations)
    for location in locations:
        x = float(location.split(".csv.gz")[0].split("_")[1])
        structure_index = int(location.split(".csv.gz")[0].split("_")[2])
        print(x, structure_index)
        df = pd.read_csv(location, compression="gzip")
        df["comp"] = x
        df["structure_index"] = structure_index

        data.append(df)

    df_merged = pd.concat(data, axis=0)
    df_merged.index.rename("n_pass", inplace=True)
    df_merged.reset_index(inplace=True)
    # df_matrices = df_merged[['comp','structure_index','time','hop_count','displacement','ekra','occupation']]
    #    df_merged.drop(['hop_count','displacement','ekra','occupation'],axis=1,inplace=True)
    return df_merged


if __name__ == "__main__":
    df = gather_data("results*.csv.gz")
    print(df.columns)
    print(df)
    df.to_csv("gathered_results.csv.gz", compression="gzip")
    # print(df.D_J)



"""
This is inherited from pymatgen.core.structure
"""

from pymatgen.core import Structure
import os
import warnings
import numpy as np
from pymatgen.core.lattice import Lattice
from pymatgen.core.sites import PeriodicSite
from pymatgen.util.coord import lattice_points_in_supercell
from monty.io import zopen


class StructureKMCpy(Structure):
    @classmethod
    def from_cif(
        cls, filename, primitive=False, sort=False, merge_tol=0.0
    ) -> "StructureKMCpy":
        """this is a modified version of Istructure/Structure.from_file which only take cif input. What change here is that, I will add a tuple, (atom_Label,) in format (str,int) as site_properties. This is for the KMCPY usage. This from_cif will call a seperated cifparser function, which is get_labeled_structure

        Args:
            filename (str): filename path
            primitive (bool, optional): not implemented?. Defaults to False.
            sort (bool, optional): not implemented?. Defaults to False.
            merge_tol (float, optional): not implemented?. Defaults to 0.0.

        Returns:
            kmcpy.external.structure.StructureKMCpy: a structure generated by the cif with all information you need.
        """
        # modified version of from file, only read the cif
        filename = str(filename)

        from kmcpy.external.cif import CifParserKMCpy

        fname = os.path.basename(filename)
        with zopen(filename, "rt") as f:
            contents = f.read()
            parser = CifParserKMCpy.from_string(contents)
            s = parser.get_labeled_structures(primitive=primitive)[0]
        return cls.from_sites(s)

    @classmethod
    def from_sites(
        cls,
        sites: list[PeriodicSite],
        charge: float | None = None,
        validate_proximity: bool = False,
        to_unit_cell: bool = False,
        properties=None,
    ) -> "StructureKMCpy":
        """
        Convenience constructor to make a StructureKMCpy from a list of sites.

        Args:
            sites: Sequence of PeriodicSites. Sites must have the same
                lattice.
            charge: Charge of structure.
            validate_proximity (bool): Whether to check if there are sites
                that are less than 0.01 Ang apart. Defaults to False.
            to_unit_cell (bool): Whether to translate sites into the unit
                cell.

        Returns:
            (StructureKMCpy) Note that missing properties are set as None.
        """
        if len(sites) < 1:
            raise ValueError(f"You need at least one site to construct a {cls}")
        prop_keys: list[str] = []
        props = properties or {}
        lattice = sites[0].lattice
        for i, site in enumerate(sites):
            if site.lattice != lattice:
                raise ValueError("Sites must belong to the same lattice")
            for k, v in site.properties.items():
                if k not in prop_keys:
                    prop_keys.append(k)
                    props[k] = [None] * len(sites)
                props[k][i] = v
        for k, v in props.items():
            if any(vv is None for vv in v):
                warnings.warn(
                    f"Not all sites have property {k}. Missing values are set to None."
                )
        return cls(
            lattice,
            [site.species for site in sites],
            [site.frac_coords for site in sites],
            charge=charge,
            site_properties=props,
            validate_proximity=validate_proximity,
            to_unit_cell=to_unit_cell,
        )

    def find_site_by_wyckoff_sequence_and_label(self, wyckoff_sequence=0, label="Na2"):
        """Look for the site with specific wyckoff sequence and label. Return the First index of site

        KMCPY function


        Args:
            wyckoff_sequence (int, optional): wyckoff sequence as defined. Defaults to 0.
            label (str, optional): label of the site. Defaults to "Na2".

        Raises:
            ValueError: if wyckoff sequence not found, means that this is not generated by from_cif
            ValueError: _description_

        Returns:
            int: index of site
        """
        if "wyckoff_sequence" not in self.site_properties:
            raise ValueError(
                "wyckoff_sequence not found in site properties.make sure that this structure is generated by kmcpy.StructureKMCpy.from_cif() function!"
            )
        for i in self._sites:
            if (
                i.properties["wyckoff_sequence"] == wyckoff_sequence
                and i.properties["label"] == label
            ):
                return i
        raise ValueError(
            "find_site_by_wyckoff_sequence_and_label:not found. with input parameter: Wyckoff sequence=",
            wyckoff_sequence,
            " label=",
            label,
        )

    def find_site_by_wyckoff_sequence_label_and_supercell(
        self, wyckoff_sequence=0, label="Na2", supercell=(0, 1, 1), return_index=True
    ):
        """Look for the site with specific wyckoff sequence and label and supercell. Return the First index of site (this should be unique)

        KMCPY function

        Args:
            wyckoff_sequence (int, optional): wyckoff sequence as defined. Defaults to 0.
            label (str, optional): label of the site. Defaults to "Na2".
            supercell (tuple): tuple of the supercell. Must be tuple with integer element!
            return_index(bool): if True, return the index of site, if False, return the site object

        Raises:
            ValueError: if wyckoff sequence not found, means that this is not generated by from_cif
            ValueError: _description_

        Returns:
            int: index of site
            or
            pymatgen.core.site
        """
        if "wyckoff_sequence" not in self.site_properties:
            raise ValueError(
                "wyckoff_sequence not found in site properties.make sure that this structure is generated by kmcpy.external.structure.StructureKMCpy.from_cif() function!"
            )
        if return_index:
            for i in range(0, len(self._sites)):
                if (
                    self._sites[i].properties["wyckoff_sequence"] == wyckoff_sequence
                    and self._sites[i].properties["label"] == label
                    and self._sites[i].properties["supercell"] == supercell
                ):
                    return i
        else:
            for i in self._sites:
                if (
                    i.properties["wyckoff_sequence"] == wyckoff_sequence
                    and i.properties["label"] == label
                    and i.properties["supercell"] == supercell
                ):
                    return i
        raise ValueError(
            "find_site_by_wyckoff_sequence_label_and_supercell:not found. with input parameter: Wyckoff sequence=",
            wyckoff_sequence,
            " label=",
            label,
            " supercell",
            supercell,
        )

    def make_kmc_supercell(
        self, scaling_matrix=(1, 2, 3), to_unit_cell=True
    ) -> "StructureKMCpy":
        """kmcpy function, nothing strange, just add a property to show which supercell this site belongs to

        Args:
            scaling_matrix (tuple, optional): scaling matrix . Defaults to (1,2,3).

        Returns:
            Structure: kmcpy.external.structure.StructureKMCpy
        """

        scale_matrix = np.array(scaling_matrix, int)
        if scale_matrix.shape != (3, 3):
            scale_matrix = np.array(scale_matrix * np.eye(3), int)
        new_lattice = Lattice(np.dot(scale_matrix, self._lattice.matrix))

        fractional_translations_for_supercell = lattice_points_in_supercell(
            scale_matrix
        )

        # c_lat = new_lattice.get_cartesian_coords(f_lat)

        # print("flat",fractional_translations_for_supercell,tuple(np.dot(fractional_translations_for_supercell,scale_matrix).astype(int)))#debug

        new_sites = []

        for site in self:

            for (
                fractional_translation_for_supercell
            ) in fractional_translations_for_supercell:

                this_properties = site.properties.copy()
                this_properties["supercell"] = tuple(
                    np.rint(
                        np.dot(fractional_translation_for_supercell, scale_matrix)
                    ).astype(int)
                )
                # print(this_properties,new_lattice.get_cartesian_coords(fractional_translation_for_supercell))
                s = PeriodicSite(
                    site.species,
                    site.coords
                    + new_lattice.get_cartesian_coords(
                        fractional_translation_for_supercell
                    ),
                    new_lattice,
                    properties=this_properties,
                    coords_are_cartesian=True,
                    to_unit_cell=True,
                    skip_checks=True,
                )

                new_sites.append(s)

        new_charge = (
            self._charge * np.linalg.det(scale_matrix) if self._charge else None
        )

        supercell = StructureKMCpy.from_sites(
            new_sites, charge=new_charge, to_unit_cell=to_unit_cell
        )

        if to_unit_cell:
            for site in supercell:
                site.to_unit_cell(in_place=True)

        return supercell

    def kmc_info_to_tuple(self, supercell=(1, 2, 3), label="Na2", wyckoff_sequence=2):
        """
        this is also a KMC specified function. Convert the information of supercell, label, wyckoff sequence into a tuple .This tuple is working as the key to the dictionary from kmc_build_dict. Mainly for accelerating the searching progress

        Args:
            supercell (tuple, optional): belongs to which supercell. Defaults to (1,2,3).
            label (str, optional): label of the atom. Defaults to "Na2".
            wyckoff_sequence (int, optional): wyckoff sequence. Defaults to 2.

        Returns:
            tuple: len=5 tuple
        """
        tmp_list = [supercell[0], supercell[1], supercell[2], label, wyckoff_sequence]
        return tuple(tmp_list)

    def site_index_vector(self, supercell=(1, 2, 3), label="Na2", local_index=2):
        """
        this is also a KMC specified function. Convert the information of supercell, label, local_index into a tuple .This tuple is working as the key to the dictionary from kmc_build_dict. Mainly for accelerating the searching progress

        Used in the version3 of generate_events

        Args:
            supercell (tuple, optional): belongs to which supercell. Defaults to (1,2,3).
            label (str, optional): label of the atom. Defaults to "Na2".
            local_index (int, optional): local_index as identifier. Defaults to 2.

        Returns:
            tuple: len=5 tuple
        """
        tmp_list = [supercell[0], supercell[1], supercell[2], label, local_index]
        return tuple(tmp_list)

    def kmc_build_dict(self, skip_check=True):
        """build a dictionary. This is a KMC specified function_summary_

        Args:
            skip_check (bool, optional): whether to skip the check on whether there is duplicate sites. Defaults to True.

        Raises:
            KeyError: if structure is not generated by KMC compatible structure.from_cif
            KeyError: if there is duplicate sites (this shouldn't happen)

        Returns:
            dict: a dictionary. Key is the tuple with same format as class.kmc_info_to_tuple, Value is the global indices
        """
        for prerequisite in ["supercell", "label", "wyckoff_sequence"]:
            if prerequisite not in self.site_properties:
                raise KeyError(
                    "make sure the structure is generated by structure.from_cif().make_kmc_supercell!"
                )
        indices_dict_from_identifier = {}

        if skip_check:

            for site_index in range(0, len(self._sites)):
                indices_dict_from_identifier[
                    self.kmc_info_to_tuple(
                        supercell=self._sites[site_index].properties["supercell"],
                        label=self._sites[site_index].properties["label"],
                        wyckoff_sequence=self._sites[site_index].properties[
                            "wyckoff_sequence"
                        ],
                    )
                ] = site_index
        else:
            for site_index in range(0, len(self._sites)):
                tmp_key = self.kmc_info_to_tuple(
                    supercell=self._sites[site_index].properties["supercell"],
                    label=self._sites[site_index].properties["label"],
                    wyckoff_sequence=self._sites[site_index].properties[
                        "wyckoff_sequence"
                    ],
                )
                if tmp_key in indices_dict_from_identifier:
                    raise KeyError(
                        "duplicate sites identifies by kmc_build_dict function. This shouldn't happen. Please check this site: ",
                        self._sites.properties,
                    )
                indices_dict_from_identifier[
                    self.kmc_info_to_tuple(
                        supercell=self._sites[site_index].properties["supercell"],
                        label=self._sites[site_index].properties["label"],
                        wyckoff_sequence=self._sites[site_index].properties[
                            "wyckoff_sequence"
                        ],
                    )
                ] = site_index

        return indices_dict_from_identifier

    def kmc_build_dict3(self, skip_check=True):
        """build a dictionary. This is a KMC specified function

        Change from the original kmc_build_dict, use the local_index as the identifier

        Args:
            skip_check (bool, optional): whether to skip the check on whether there is duplicate sites. Defaults to True.

        Raises:
            KeyError: if structure is not generated by KMC compatible structure.from_cif
            KeyError: if there is duplicate sites (this shouldn't happen)

        Returns:
            dict: a dictionary. Key is the tuple with same format as class.kmc_info_to_tuple, Value is the global indices
        """
        for prerequisite in ["supercell", "label", "local_index"]:
            if prerequisite not in self.site_properties:
                raise KeyError(
                    "make sure the structure is generated by structure.from_cif().make_kmc_supercell!"
                )

        indices_dict_from_identifier = {}

        if skip_check:

            for site_index in range(0, len(self._sites)):
                indices_dict_from_identifier[
                    self.site_index_vector(
                        supercell=self._sites[site_index].properties["supercell"],
                        label=self._sites[site_index].properties["label"],
                        local_index=self._sites[site_index].properties["local_index"],
                    )
                ] = site_index
        else:
            for site_index in range(0, len(self._sites)):
                tmp_key = self.site_index_vector(
                    supercell=self._sites[site_index].properties["supercell"],
                    label=self._sites[site_index].properties["label"],
                    local_index=self._sites[site_index].properties["local_index"],
                )

                if tmp_key in indices_dict_from_identifier:
                    raise KeyError(
                        "duplicate sites identifies by kmc_build_dict function. This shouldn't happen. Please check this site: ",
                        self._sites.properties,
                    )

                indices_dict_from_identifier[
                    self.site_index_vector(
                        supercell=self._sites[site_index].properties["supercell"],
                        label=self._sites[site_index].properties["label"],
                        local_index=self._sites[site_index].properties["local_index"],
                    )
                ] = site_index

        return indices_dict_from_identifier



#!/usr/bin/env python
import pandas as pd
import numpy as np
import glob
import os
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick


results = glob.glob("comp_0.1_1")


def calc_D_J(displacement, n_na, time, d=3):
    displacement_vector_tot = np.linalg.norm(np.sum(displacement, axis=0))
    D_J = displacement_vector_tot**2 / (2 * d * time * n_na) * 10 ** (-16)  # to cm^2/s
    print("log10(10^13/v* D_J) = ", np.log10(D_J), "cm^2/s")
    print("D_J = ", D_J, "cm^2/s")
    return D_J


def calc_D_tracer(displacement, n_na, time, d=3):
    displacement_vector_tot = np.sum((np.linalg.norm(displacement, axis=1)) ** 2)
    D_tracer = (
        displacement_vector_tot / (2 * d * time * n_na) * 10 ** (-16)
    )  # to cm^2/s
    print("log10(10^13/v* D_tracer) =", np.log10(D_tracer), "cm^2/s")
    print("D_tracer =", D_tracer, "cm^2/s")
    return D_tracer


def calc_corr_factor(
    displacement, n_na, hop_counter, a=3.47782
):  # a is the hopping distance in Angstrom
    n = np.average(hop_counter)
    # distance_squared_avg = np.average((self.hop_counter*a)**2)
    corr_factor = np.sum(np.linalg.norm(displacement) ** 2) / (n_na * n * a**2)
    # corr_factor = distance_squared_avg/(self.n_na*n*a**2)
    print("Correlation factor f = ", corr_factor)
    return corr_factor


def calc_D_c(D_J, theta):
    D_c = D_J * theta
    return D_c


def fit_theta():
    x, y = np.loadtxt("theta_575.txt")  # x,theta
    from scipy.interpolate import interp1d

    spl = interp1d(x, y, kind="cubic")
    return spl


def calc_conductivity(D_J, D_tracer, n_na, volume, q=1, T=300):
    H_R = D_tracer / D_J
    n = (n_na) / volume  # e per Angst^3 vacancy is the carrier
    k = 8.617333262145 * 10 ** (-2)  # unit in meV/K
    conductivity = D_J * n * q**2 / (k * T) * 1.602 * 10**11  # to mS/cm
    print("Haven's ratio H_R =", H_R)
    print("Conductivty: sigma = ", conductivity, "mS/cm")
    return conductivity


def search_string(fname, word, location):
    with open(fname) as f:
        for l in f.readlines():
            if word in l:
                res = l.strip().split()[location]
    return float(res)


spl = fit_theta()
data = []
for result in results:
    for n_pass in np.arange(0, 2000, 1, dtype="int"):
        print(result)
        x = 3 - 3 * float(result.strip().split("_")[1])
        structure_id = int(result.strip().split("_")[2])
        D_J, D_tracer, f, conductivity = np.loadtxt(
            result + "/migration_" + str(n_pass) + ".txt"
        )
        # displacement = np.loadtxt(result+'/displacement.txt')
        # hop_counter = np.loadtxt(result+'/hop_count.txt')
        # time = search_string(result+'/kmc_stdout.txt','Time',2)
        # n_na = search_string(result+'/kmc_stdout.txt','n_Na_sites',2)
        # volume = 2213256.28246698
        # D_J = calc_D_J(displacement,n_na,time,d=3)
        # D_tracer = calc_D_tracer(displacement,n_na,time,d=3)
        theta = spl(x)
        D_c = calc_D_c(D_J, theta)
        print("D_c = ", D_c)
        # f = calc_corr_factor(displacement,n_na,hop_counter)
        # conductivity = calc_conductivity(D_J,D_tracer,n_na,volume,q=1,T=572)
        data.append(
            [
                x,
                n_pass,
                structure_id,
                D_J,
                D_tracer,
                D_c,
                theta,
                f,
                conductivity,
                result,
            ]
        )


if __name__ == "__main__":
    df = pd.DataFrame(
        data,
        columns=[
            "x",
            "n_pass",
            "structure_id",
            "D_J",
            "D_tracer",
            "D_c",
            "Theta",
            "f",
            "Conductivity",
            "location",
        ],
    ).sort_values(by=["x", "n_pass"])

    df.to_csv("data_all.csv")
    df2 = df.groupby("x").mean().reset_index().sort_values(by="x")
    print(df2)
    df2.to_csv("data.csv")

    df2 = pd.read_csv("data.csv")
    fig, axes = plt.subplots(2, 2, figsize=(5, 4), sharex="col")
    axes[0, 0].plot(df2.x, df2.D_J, ".")
    axes[0, 0].set_ylabel(r"$D_{J}$ (cm$^2$/s)")
    axes[0, 0].yaxis.set_major_formatter(mtick.FormatStrFormatter("%.2e"))

    axes[0, 1].plot(df2.x, df2.D_tracer, ".")
    axes[0, 1].set_ylabel(r"$D^{*}$ (cm$^2$/s)")
    axes[0, 1].yaxis.set_major_formatter(mtick.FormatStrFormatter("%.2e"))

    axes[1, 0].plot(df2.x, df2.f, ".")
    axes[1, 0].set_ylabel(r"Correlation factor")
    axes[1, 0].set_xlabel(
        r"${\rm x}$ in Na$_{\rm 1+x}$Zr$_{\rm 2}$Si$_{\rm x}$P$_{\rm 3-x}$O$_{\rm 12}$"
    )
    axes[1, 0].set_ylim((0, 1))
    axes[1, 0].set_xlim((0, 3))

    axes[1, 1].plot(df2.x, df2.Conductivity, ".")
    axes[1, 1].set_ylabel(r"$\sigma$ (mS/cm)")
    axes[1, 1].set_xlabel(
        r"${\rm x}$ in Na$_{\rm 1+x}$Zr$_{\rm 2}$Si$_{\rm x}$P$_{\rm 3-x}$O$_{\rm 12}$"
    )
    axes[1, 1].set_xlim((0, 3))

    fig.tight_layout()

    fig.savefig("plot.pdf")



"""
This is inherited from pymatgen.analysis.local_env
"""

from pymatgen.analysis.local_env import CutOffDictNN, NearNeighbors
from kmcpy.external.structure import StructureKMCpy
from monty.serialization import loadfn
import os

_directory = os.path.join(os.path.dirname(__file__))


class CutOffDictNNKMCpy(CutOffDictNN):
    def get_nn_info(self, structure, n):
        """
        Get all near-neighbor sites as well as the associated image locations
        and weights of the site with index n in structure.

        Args:
            structure (Structure): input structure.
            n (integer): index of site for which to determine near-neighbor
                sites.

        Returns:
            siw (list of tuples (Site, array, float)): tuples, each one
                of which represents a coordinated site, its image location,
                and its weight.
        """
        site = structure[n]

        neighs_dists = structure.get_neighbors(site, self._max_dist)

        nn_info = []

        if "wyckoff_sequence" in structure.site_properties:
            if "supercell" in structure.site_properties:
                for nn in neighs_dists:
                    n_site = nn
                    dist = nn.nn_distance
                    neigh_cut_off_dist = self._lookup_dict.get(
                        site.species_string, {}
                    ).get(n_site.species_string, 0.0)

                    if dist < neigh_cut_off_dist:
                        nn_info.append(
                            {
                                "site": n_site,
                                "image": self._get_image(structure, n_site),
                                "weight": dist,
                                "site_index": self._get_original_site(
                                    structure, n_site
                                ),
                                "wyckoff_sequence": n_site.properties[
                                    "wyckoff_sequence"
                                ],
                                "local_index": n_site.properties["local_index"],
                                "label": n_site.properties["label"],
                                "supercell": n_site.properties["supercell"],
                            }
                        )
            else:
                for nn in neighs_dists:
                    n_site = nn
                    dist = nn.nn_distance
                    neigh_cut_off_dist = self._lookup_dict.get(
                        site.species_string, {}
                    ).get(n_site.species_string, 0.0)

                    if dist < neigh_cut_off_dist:
                        nn_info.append(
                            {
                                "site": n_site,
                                "image": self._get_image(structure, n_site),
                                "weight": dist,
                                "site_index": self._get_original_site(
                                    structure, n_site
                                ),
                                "wyckoff_sequence": n_site.properties[
                                    "wyckoff_sequence"
                                ],
                                "local_index": n_site.properties["local_index"],
                                "label": n_site.properties["label"],
                            }
                        )
        else:
            for nn in neighs_dists:
                n_site = nn
                dist = nn.nn_distance
                neigh_cut_off_dist = self._lookup_dict.get(site.species_string, {}).get(
                    n_site.species_string, 0.0
                )

                if dist < neigh_cut_off_dist:
                    nn_info.append(
                        {
                            "site": n_site,
                            "image": self._get_image(structure, n_site),
                            "weight": dist,
                            "site_index": self._get_original_site(structure, n_site),
                        }
                    )

        return nn_info


class CutOffDictNNrange(NearNeighbors):
    """
    Jerry: Modified from CutOffDictNN in pymatgen so that it can search all pairs between a range of distances [d_min, d_max]
    c
    A basic NN class using a dictionary of fixed cut-off distances.
    Only pairs of elements listed in the cut-off dictionary are considered
    during construction of the neighbor lists.
    Omit passing a dictionary for a Null/Empty NN class.
    """

    def __init__(self, cut_off_dict=None):
        """
        Args:
            cut_off_dict (dict[str, float]): a dictionary
            of cut-off distances, e.g. {('Fe','O'): [2.0,3.0]} for
            a maximum Fe-O bond length between 2.0 and 3.0 Angstroms.
            Bonds will only be created between pairs listed
            in the cut-off dictionary.
            If your structure is oxidation state decorated,
            the cut-off distances will have to explicitly include
            the oxidation state, e.g. {('Fe2+', 'O2-'): [2.0,3.0]}
        """
        from collections import defaultdict

        self.cut_off_dict = cut_off_dict or {}
        # for convenience  Jerry: added minimum and maximum
        self._max_dist = 0.0
        self._min_dist = 1e3
        lookup_dict_max = defaultdict(dict)
        lookup_dict_min = defaultdict(dict)
        for (sp1, sp2), (dist_min, dist_max) in self.cut_off_dict.items():
            lookup_dict_max[sp1][sp2] = dist_max
            lookup_dict_max[sp2][sp1] = dist_max
            if dist_max > self._max_dist:
                self._max_dist = dist_max

            lookup_dict_min[sp1][sp2] = dist_min
            lookup_dict_min[sp2][sp1] = dist_min
            if dist_min < self._min_dist:
                self._min_dist = dist_min
        self._lookup_dict_max = lookup_dict_max
        self._lookup_dict_min = lookup_dict_min

    @property
    def structures_allowed(self):
        """
        Boolean property: can this NearNeighbors class be used with Structure
        objects?
        """
        return True

    @property
    def molecules_allowed(self):
        """
        Boolean property: can this NearNeighbors class be used with Molecule
        objects?
        """
        return True

    @property
    def extend_structure_molecules(self):
        """
        Boolean property: Do Molecules need to be converted to Structures to use
        this NearNeighbors class? Note: this property is not defined for classes
        for which molecules_allowed is False.
        """
        return True

    @staticmethod
    def from_preset(preset):
        """
        Initialise a CutOffDictNN according to a preset set of cut-offs.
        Args:
            preset (str): A preset name. The list of supported presets are:
                - "vesta_2019": The distance cut-offs used by the VESTA
                  visualisation program.
        Returns:
            A CutOffDictNN using the preset cut-off dictionary.
        """
        if preset == "vesta_2019":
            cut_offs = loadfn(os.path.join(_directory, "vesta_cutoffs.yaml"))
            return CutOffDictNNrange(cut_off_dict=cut_offs)

        raise ValueError(f"Unrecognised preset: {preset}")

    def get_nn_info(self, structure: StructureKMCpy, n: int):
        """
        Get all near-neighbor sites as well as the associated image locations
        and weights of the site with index n in structure.
        Args:
            structure (StructureKMCpy): input structure.
            n (int): index of site for which to determine near-neighbor
                sites.
        Returns:
            siw (list of tuples (Site, array, float)): tuples, each one
                of which represents a coordinated site, its image location,
                and its weight.
        """
        site = structure[n]

        neighs_dists = structure.get_neighbors(site, self._max_dist)

        nn_info = []
        for nn in neighs_dists:
            n_site = nn
            dist = nn.nn_distance
            neigh_cut_off_dist_max = self._lookup_dict_max.get(
                site.species_string, {}
            ).get(n_site.species_string, 0.0)
            neigh_cut_off_dist_min = self._lookup_dict_min.get(
                site.species_string, {}
            ).get(n_site.species_string, 0.0)
            if dist < neigh_cut_off_dist_max and dist > neigh_cut_off_dist_min:
                nn_info.append(
                    {
                        "site": n_site,
                        "image": self._get_image(structure, n_site),
                        "weight": dist,
                        "site_index": self._get_original_site(structure, n_site),
                    }
                )

        return nn_info
